#! /bin/sh
cat <<EOF
################################################################################
#                  _______        _    _____       _ _                         # 
#                 |__   __|      | |  / ____|     (_) |                        # 
#                    | | ___  ___| |_| (___  _   _ _| |_ ___                   #
#                    | |/ _ \/ __| __|\___ \| | | | | __/ _ \                  #
#                    | |  __/\__ \ |_ ____) | |_| | | ||  __/                  # 
#                    |_|\___||___/\__|_____/ \__,_|_|\__\___|                  # 
#                                                                              #
#                 written by Sergei Ivanov and Gerald Mathias                  #
#                                                                              #
#     Send comments, suggestions and complaints regarding the TestSuite to     #
#                                                                              #
#                  sergei.ivanov@theochem.ruhr-uni-bochum.de                   #
#                                  or                                          #
#                  gerald.mathias@theochem.ruhr-uni-bochum.de                  #
#                                                                              #
################################################################################

EOF
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#!!         PLEASE, DON'T CHANGE THE SCRIPT, except for porting to            !!
#!!      different platforms (proper variables for the utilities for this     !!
#!!      purpose can be found few lines below). The Test Suite has enough     !!
#!!      options to satisfy ALMOST ANY NEEDS. The manual is in the file       !!
#!!      'Description.tex' placed in the same directiory, as the script -     !!
#!!                              READ IT FIRST  !!!                           !!
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

###################################################################################
#                              PREAMBLE                                           #
###################################################################################
export LC_ALL=C

# Filenames
ERR_LOG="Errors.log"
OUT_LOG="Tests.log"
MAKE_LOG="make.log"
SUPPRESS="no_msg.log"
INFO_LOG="Info.log"
TEST_LIST="file.tests"
FAIL_LIST="failed.tests"
CPMD_FLIST="outfiles.list"
SUITE_FLIST="suitefiles.list"
BASE_DIR="${PWD}"

# no default tests
TESTDIRS=""

# utility aliases for easier porting to different platforms
AWK='awk'
myCC='cc'
CAT='cat'
CP='cp'
CUT='cut'
FIND='find'
GREP='grep'
HEAD='head'
MAKE='make'
MV='mv'
RM='rm'
SORT='sort'
SED='sed'
TAIL='tail'
UNIQ='uniq'
MD5='md5sum'
WC='wc'
TR='tr'
PASTE='paste'

#platform dependent definitions
case $(uname) in
  SunOS) datetr=$(date);
         DIFF='diff -r'
         ;;
  Linux) mydate=$(date --iso-8601="seconds");
         datetr=$(echo ${mydate} | ${TR} 'T' ' ');
         # patchfile like differences
         DIFF='diff -Naur'
         ;;
  AIX  ) datetr=$(date);
         DIFF='diff -r'
         ;;
   *   ) echo "The system is not recognized. Uses Default."
         datetr=`date`;
         DIFF='diff -r'
         ;; 
esac

###################################################################################
#                   __                  _   _                                     #
#                  / _|_   _ _ __   ___| |_(_) ___  _ __  ___                     #
#                 | |_| | | | '_ \ / __| __| |/ _ \| '_ \/ __|                    #
#                 |  _| |_| | | | | (__| |_| | (_) | | | \__ \                    #  
#                 |_|  \__,_|_| |_|\___|\__|_|\___/|_| |_|___/                    #
#                                                                                 #
###################################################################################

###################################################################################
#   extract field no ${1} from '|' separated (alt. ${3} separated) string ${2}    #
###################################################################################
 GetField()
{
if [ ${#} -eq 3 ] ; then
  d="${3}"
else
  d='|'
fi
# only the first field may correspond to the whole line (see man cut)
if [ ${1} -gt 1 ] ; then
   field="$(echo "${2}" | ${CUT} -s -f${1} -d${d})" 
else
   field="$(echo "${2}" | ${CUT}    -f${1} -d${d})" 
fi 
echo "${field}"
}

###################################################################################
#          return true/false boolean function depending on keyword                #
###################################################################################
GetBool()
{
key=${1}
if [ "${key:1:2}" = "no" ] ; then
   echo "false"
else
   echo "true"
fi
}

###################################################################################
#                Setting the default value loaded from the defaults.cfg           #
###################################################################################
set_default()
{
 VAR=${1}
 line=`${AWK} "/^${VAR}\=/ {print} " ${DEF}`
 DEFAULT="$(GetField 2 "${line}" '=')"
 echo "${DEFAULT}"
}

###################################################################################
#                   load the default values from the file defaults.cfg            #
###################################################################################
Load_Defaults()
{
 DEF="${BASE_DIR}/defaults.cfg"
 if [ ! -s "${DEF}" ]; then 
     end_test 11 "!!! The defaults file ${DEF} is absent !!!"; 
 fi
         vc="$(set_default vc)";        
        run="$(set_default run)";       
      clean="$(set_default clean)";     
      vcupd="$(set_default vcupd)";   
       vcig="$(set_default vcig)";     
    compile="$(set_default compile)";    
     reconf="$(set_default reconf)";    
      LEVEL="$(set_default LEVEL)";     
       CPMD="$(set_default CPMD)";
     endian="$(set_default endian)";     
      accur="$(set_default accur)";     
        add="$(set_default add)";       
     locdef="$(set_default locdef)";
       wipe="false";       
    MAXPROC="$(set_default MAXPROC)";   
  maxdigits="$(set_default maxdigits)"; 
   ALLTESTS="$(set_default ALLTESTS)";
}

###################################################################################
#      spot logical flaws after defaults and cmd line have been parsed            #
###################################################################################
Check_Input()
{
  if [ "${numeric:-XXX}" == "numeric" -a "${literal:-XXX}" == "literal" ] ; then
     end_test 12 "!!! -numeric and -literal are mutually exclusive !!!" 
  fi
  case ${vc} in
     cvs|svn) ;;
     none) 
       # no version control  no update
       vcupd="false"
       ;;
     *) end_test 12 "!!! unsupported version control <${vc}> !!!" 
     ;;
  esac
  if [ $(( (${LEVEL} < 1) || (${LEVEL} > 3) )) == 1 ]; then
     end_test 12 'Level must be in the interval [1:3]'
  fi;
  ## place other checks here
  if  ${wipe} ; then
     run=false
     add=false
  fi
}

###################################################################################
#          set of simple functions to output to the stdout and to a file          #
###################################################################################
# output
 outecho(){
 echo "${@}" >> ${OUT_DIR}/${OUT_LOG}
 echo "${@}"
}
###################################################################################
# errors
 errecho()
{
 echo "${@}" >> ${OUT_DIR}/${ERR_LOG}
# echo "${@}" 
}
###################################################################################
# formatted output:
# write output to ${1} 10 character columns
# negative ${1} writes left adusted
 outecho_column()
{
 pos=${1}0
 trunc=$(( pos > 0 ? pos : -pos ))
 printf "%${pos}s" "${2:0:${trunc}}" >> ${OUT_DIR}/${OUT_LOG}
 printf "%${pos}s" "${2:0:${trunc}}"
}
###################################################################################
# terminate an output line
 outecho_status()
{
 pos="10"
 printf "%${pos}s\n" "${@}" >> ${OUT_DIR}/${OUT_LOG}
 printf "%${pos}s\n" "${@}"
}

###################################################################################
# for error output
 errecho_column()
{
 pos=${1}0
 shift
 printf "%${pos}s" "${@}" >> ${OUT_DIR}/${ERR_LOG}
# printf "%${pos}s" "${@}"
}
###################################################################################
# an output separator
 hline()
{
 outecho "--------------------------------------------------------------------------------"
}

###################################################################################
# generate a centered char-text-char line
string_separator ()
{
local str="${1}"
local char
if [ ${#} -lt 2 ] ; then
   char='-'
else
   char="${2}"
fi

 echo | ${AWK} "{c=\"${char}\"; full=c \"${str}\" c; while(length(full)<80) full=c full c ; printf(\"%s\n\",substr(full,0,79))} "
}
###################################################################################
#                exit function to end the tests all in the same way               #
###################################################################################
 end_test() 
{
hline
clean_suitefiles GLOBAL
outecho ${2}
outecho_column -5 "Testing finished" 
outecho_column 3 "$(date)"
outecho
hline
hline
exit ${1}
}

###################################################################################
#                 Test whether the directory exists or not                        #
###################################################################################
 Dir_Exist()
{
 if [ ! -d "${1}" ]; then 
   case ${2} in
    CREATE)  mkdir -p ${1};
             if (( ${?} )); then
               echo "!!!  Can't create directory ${1}  !!!" >> ${OUT_DIR}/${ERR_LOG} 2> /dev/null
               end_test 5 "!!!  Can't create directory ${1}  !!!"
             else
               outecho "${3}: ${1} is created" 2> /dev/null
             fi;;
    EXIT)    end_test 6 "${3}: ${1} does not exist";;
       *)    end_test 7 "Wrong option for Dir_Exist for ${1} with argument: ${2}";;
   esac
 else
   outecho_column  -2 "${3}:" 
   outecho_column  -6 "$(echo ${1}|${SED} s@${HOME}@\~@)"
   outecho 
 fi
}
###################################################################################
#                              expand relative pathes                             #
###################################################################################
 absolute_path ()
{
  if [ -d ${1} ] ; then
    mydir=${PWD}
    cd ${1}
    echo ${PWD}
    cd ${mydir}
  else
    echo ${1}
  fi
 return 0
}

###################################################################################
#                                   backup old log files                          #
###################################################################################
 rotate_log () 
{
  BASE=${1}
  test -f ${OUT_DIR}/${BASE} || return 0
  nextnum=$( ${FIND} ${OUT_DIR} -name "\#${BASE}*" |\
             ${SED} "s@.*#${BASE}.\([0-9]*\).*@\1@" |\
             ${SORT} -n | ${TAIL} -1)
  nextnum=$((${nextnum}+1))
  ${MV} ${OUT_DIR}/${BASE} ${OUT_DIR}/\#${BASE}.${nextnum}\#
  printf "%-40s%-30s%10s\n" "Backing up ${BASE}" "to #${BASE}.${nextnum}#" "DONE"
  return 0
}

###################################################################################
#              clean files generated by testsuite of type ${tag}                  #
###################################################################################
clean_suitefiles()
{
local tag=${1}
local snum=${2}
for file in $(${GREP} ${tag} ${BASE_DIR}/${SUITE_FLIST}) ; do
    del=$(GetField 1 ${file})
    ${RM} -rf ${del} ${snum}.${del} 2>> ${OUT_DIR}/${SUPPRESS}
done
}

###################################################################################
#                          clean the test directory                               #
###################################################################################
 clean_run()
{
local snum=${1}
for file in $(${CAT} ${BASE_DIR}/${CPMD_FLIST}|${CUT} -f1 -d'|') ; do
    ${RM} -rf ${file} 2>> ${OUT_DIR}/${SUPPRESS}
    ${RM} -rf ${snum}.${file} 2>> ${OUT_DIR}/${SUPPRESS}
done
clean_suitefiles RUN ${snum}
}

###################################################################################
#                         save output files for this variant                      #
###################################################################################
 CopyRunOutput ()
{
 rn="${1}"
 for file in $(${CAT} ${BASE_DIR}/${CPMD_FLIST}|${CUT} -f1 -d'|') ; do
  if [ -e "${file}" ]; then 
    ${MV} ${file} ${rn}.${file}
  fi
 done
 for file in $(${GREP} RUN ${BASE_DIR}/${SUITE_FLIST}|${CUT} -f1 -d'|') ; do
  if [ -e "${file}" ]; then 
    ${MV} ${file} ${rn}.${file}
  fi
 done
} 

###################################################################################
#              ignore files for version contol in this test directory            #
###################################################################################
VC_ignore()
{
${AWK} -F'|' '{printf("%s\n[0-9][0-9].%s\n",$1,$1)}'\
              ${BASE_DIR}/${CPMD_FLIST} > ignore.tmp
${AWK} -F'|' '$2=="RUN" || $2=="KEYWORD" {printf("%s\n[0-9][0-9].%s\n",$1,$1)}; \
              $2=="TEST"  {printf("%s\n",$1)}' \
              ${BASE_DIR}/${SUITE_FLIST} >> ignore.tmp
# if we have to include specific ignore files for a test, we can do it here
#...
case ${vc} in
   cvs)
      ${MV} ignore.tmp .cvsignore
   ;;
   svn)
      svn propset svn:ignore -F ignore.tmp .\
           >> ${OUT_DIR}/${SUPPRESS} 2>> ${OUT_DIR}/${ERR_LOG} 
   ;;
esac
 ${RM} -rf ignore.tmp 
}

###################################################################################
#       prepare the input/RESTART etc. files of the proper level to ./input etc.  #
###################################################################################
cp_input_files()
{ 
# generate the input file
sstr="${1}"
case ${myendian} in
  LITTLE)
       edn="lit"
       ;;
  BIG)
       edn="big"
       ;;
  *)   
       errecho "unknown endianess <${myendian}>"
       end_test 13
  ;;
esac
${SED} "${sstr}" test.in.${LEVEL} > input 2>> ${OUT_DIR}/${SUPPRESS} 
# check for predefined restart files
KAR=$(ls RESTART*.${edn} 2> /dev/null | ${WC} -l) 
if [ ${KAR} -gt 0 ]; then 
   for res in RESTART*.${edn}; do
   ${CP} ${res} ${res%.${edn}}
   done
fi

if [ -e LATEST.1 ]; then 
   ${CP} LATEST.1 LATEST 2> /dev/null;
else 
   echo -e "RESTART.1\n        1" > LATEST 2> /dev/null
fi



if [ -s GEOMETRY.gen ] && ${run} ; then 
   ${CP} GEOMETRY.gen GEOMETRY  2> /dev/null;
fi 

}

###################################################################################
#              check for default input on this test level                         #
#                    and generate a list of variants                              #
###################################################################################
gen_inputs() 
{
# generate list of input variats
if [ -e test.in.${LEVEL} ]; then
         ${RM} -f curtests
      echo "00|# no change|generic|${MAXPROC}" >> curtests
   num_tests=01    
   if [ -e VARIANTS.${LEVEL} -a "${add}" = "false" ]; then
            varcpu=$(echo ${MPICMD}|${GREP} -c NCPU)
            while read line ; do
                     minc=$(GetField 1 "${line}")
                     maxc=$(GetField 2 "${line}")
                     sstr="$(GetField 3 "${line}")"
                     cstr="$(GetField 4 "${line}")"
                     # skip empty lines
                     if [ -z "${cstr}" ] ; then
                        continue
                     fi
                     # default is to use all cpus available
                     np=${MAXPROC}
                     if ! ${wipe} ; then
                        # test for comment lines
                        if [ -z "$(echo ${minc}|${TR} -d -c '[0-9]')" ] ; then
                              continue
                        fi
                        if [ ${MAXPROC} -lt ${minc} -a  ${minc} -ne 0 ] ; then
                              continue
                        fi
                        if [ ${MAXPROC} -gt ${maxc} -a  ${maxc} -ne 0 ] ; then
                              if [ ${varcpu} -gt 0 ] ; then
                                 np=${maxc}
                              else
                                 continue
                              fi
                        fi
                        if [ -z "${sstr}" ] ; then
                            sstr="# no change"
                        fi
                     fi
                     echo "${num_tests}|${sstr}|${cstr}|${np}" >> curtests
                     num_tests=$(echo ${num_tests}|awk '{printf("%02i",$1+1)}')
            done < VARIANTS.${LEVEL}
   fi
   echo ${num_tests}
   set +vx
   return 0
fi
echo 0
set +vx
return 1
}

###################################################################################
#               Checking whether the files for comparison exist or not            #
###################################################################################
 Files_Exist()
{
while [ ${#} -ge 1 ]; do
 if [ ! -s  "${1}" ]; then 
   echo "    File ${1} does not exist !!!" >> ./errors;
 return 1;
 fi
 shift
done
return 0
}

###################################################################################
#       filter function removing minus signes from all but from exponentials      #
###################################################################################
 strip_minus ()
{
  ${SED} 's/\([^EeDd]\)-/\1 /g'
}

###################################################################################
#                     calculate the walltime spent for a test                     #
###################################################################################
 Eltime()
{ 
local s=${1}
local e=${2}
local t h m s mytime
 let t=e-s
 let h=t/3600
 let t=t%3600
 let m=t/60
 let s=t%60
 mytime=$(printf "%02i:%02i:%02i" ${h} ${m} ${s} )
 outecho_column -1 ${mytime}
}
###################################################################################
#                     date to seconds                                             #
###################################################################################
 myseconds() 
{  
   # cumbersome but SunOS proof
  echo $(date +%H_%M_%S | ${AWK} -F_ '{print $1*3600+$2*60+$3}')
}

###################################################################################
#         check binary enconding of floats (endian type )                         #
###################################################################################
 get_endian()
{
${CAT} <<EOF > test.c
int     main(void)               /*** 0 little endian ***/
{                               /*** 1 big endian    ***/
        int     test=1;
        char    *a;
        a = (char*) &test;
        if (*a==1) return 0;
        return 1;
}
EOF

${myCC} test.c -o test.x >> ${OUT_DIR}/${SUPPRESS} 2>&1
./test.x
let endian=${?}
if [ ${endian} -eq 0 ] ; then
 echo "LITTLE"
else
 echo "BIG"
fi
return ${endian}
}

###################################################################################
#                             Substitution for bc                                 #
###################################################################################
 make_myBC()
{
 ${CAT} <<EOF > bc.c
#include <stdio.h>
#include <string.h>
#include <math.h>

#define SMALL 1.E-32
int     main(void)
{
 int lines, acc,maxdigits, i, Est_acc1, Est_acc2, Est_acc, j, ACC;
 char TempStr[256], NumStr1[128], NumStr2[128];
 int Len1, Len2, Res=0;
 double Num1, Num2, Denom, Diff, Deviation;
 double mDev,vDev,mDiff,vDiff;
 int INum1, INum2;

 scanf("%d%d%d",&lines, &acc, &maxdigits);
 for(i=0;i<lines;i++) 
 {
  scanf("%s%s", NumStr1, NumStr2);
  
  if ( (strchr(NumStr1,'.')==NULL) && (strchr(NumStr2,'.')==NULL) )
  {
   sscanf(NumStr1,"%d", &INum1);
   sscanf(NumStr2,"%d", &INum2);
   if (INum1!=INum2)
   {
    printf("%s  %s\n", NumStr1, NumStr2);
    Res++;
   }
   continue;
  }

  Len1=strlen(NumStr1);
  Len2=strlen(NumStr2);

  j=0;
  Est_acc1=0;
  while( (NumStr1[j]!='E') && (j<Len1) )
  {
   if (isdigit(NumStr1[j]))
    if ((NumStr1[j]!='0') || (Est_acc1>0) ) Est_acc1++;
   j++;
  }

  j=0;
  Est_acc2=0;
  while( (NumStr2[j]!='E') && (j<Len2) )
  {
   if (isdigit(NumStr2[j]))
    if ((NumStr2[j]!='0') || (Est_acc2>0) ) Est_acc2++;
    j++;
  }

  Est_acc=(Est_acc1>Est_acc2) ? Est_acc1 : Est_acc2;
  ACC=(Est_acc>maxdigits && maxdigits > 0) ? maxdigits : Est_acc;
  ACC-=acc;
  sscanf(NumStr1,"%lf", &Num1);
  sscanf(NumStr2,"%lf", &Num2);

  Diff=-(Num1-Num2);
  Denom=0.5 *( fabs(Num1)+fabs(Num2));
  if ( (Denom<SMALL) || (fabs(Diff)<SMALL) ) continue;
  Deviation= Diff/Denom;
  if (fabs(Deviation)>pow(10.0,-ACC))
  {
   printf("%20s<>%20s| difference (rel.): %12.4e (%12.4e)\n", NumStr1,NumStr2,Diff,Deviation);
   Res++;
  }
  mDiff+=Diff;
  vDiff+=Diff*Diff;
  mDev+=Deviation;
  vDev+=Deviation*Deviation;
 } 
 if(Res>0)
 {
   printf("mean difference (stddev)         : %12.4e (%12.4e)\n",mDiff/lines,sqrt(vDiff/lines-mDiff/lines*mDiff/lines) );
   printf("mean relative difference (stddev): %12.4e (%12.4e)\n",mDev/lines, sqrt(vDev/lines -mDev/lines *mDev/lines) );
   printf("%d out of %d matches differ.\n",Res,lines);
   return 1;
 } 
 return 0;
}
EOF
 cd ${BASE_DIR}
 ${myCC} bc.c -o myBC.x -lm >> ${OUT_DIR}/${SUPPRESS} 2>&1
 if (( ${?} )); then end_test 15 "Can't compile my BC"; fi
}

###################################################################################
#                        Start serial and parallel CPMD                           #
###################################################################################
 Run_CPMD()
{ 
local INPUT=${1}
local OUTPUT=${2}
local RUNNCPU=${3}
 if [ ! -x "${EXE_DIR}/${CPMD}" -o ! -f "${EXE_DIR}/${CPMD}" ]; then 
    outecho
    outecho "Program: ${EXE_DIR}/${CPMD} is not executable" 
    end_test 10
 fi
 if [ -n "${MPICMD}" ] ; then
   # '< /dev/null' is needed to prevent mpirun from breaking file handles
   ${MPICMD/NCPU/${RUNNCPU}} ${EXE_DIR}/${CPMD} ${INPUT} ${PS_DIR} \
                        >> ${OUTPUT} 2>> ${OUT_DIR}/${SUPPRESS} < /dev/null
   let result=${?}
 else 
    ${EXE_DIR}/${CPMD} ${INPUT} ${PS_DIR} \
                        >> ${OUTPUT} 2>> ${OUT_DIR}/${SUPPRESS}
   let result=${?}
 fi
 if ((${result})) ; then 
   echo " CPMD returned ${result}" >> ./errors;
   return 1;
 fi
 return 0
}

###################################################################################
#                     Extract all numbers from given files                        #
###################################################################################
 Extract_Numbers()
{
while [ ${#} -gt 0 ]  ; do
 ${CAT} ${1} | ${TR} -cd "+\-.0123456789 \nEedD" |\
     ${TR} ' edD' "\nEEE" | ${GREP} '[0-9]' |\
     ${AWK} '/^[-+]?[0-9]*\.?[0-9]+(E[-+0-9])?[0-9]*$/ {print $0}'\
             > ${1}.num 2>>${OUT_DIR}/${SUPPRESS}
     shift
done
}

###################################################################################
#               Compare two files of one column each numerically                  #
###################################################################################
 Compare_Files_Num()
{
cfile="${1}"
rfile="${2}"
ckey="${3}"
kword="${4}"
fstring="${5}"
ktype="${6}"
 Files_Exist ${cfile}.num
 num=`    ${WC} -l ${cfile}.num    | ${AWK} '{print $1}' 2>>${OUT_DIR}/${SUPPRESS}`
 num_ref=`${WC} -l ${rfile}.num    | ${AWK} '{print $1}' 2>>${OUT_DIR}/${SUPPRESS}`
 if [ ${num} != ${num_ref} ]; then
   echo "    Numbers of values in ${ckey} ${kword} from file ${fstring}: ${num} differ from the reference: ${num_ref} !!!" >> diff.err
   append_diff "${ckey}|${kword}|${fstring}" errors 
   return 1;           
 fi;              
 
 mydigits=$(echo ${ktype}|${SED} 's/NUM//' |${AWK} -F\- '{print $1}')
 myaccur=$(echo  ${ktype}|${SED} 's/NUM//' |${AWK} -F\- '{print $2}')

 echo "${num} ${myaccur} ${mydigits}"      > ${cfile}.cmp 2>>${OUT_DIR}/${SUPPRESS}
 ${PASTE} ${cfile}.num ${rfile}.num >> ${cfile}.cmp 2>>${OUT_DIR}/${SUPPRESS}
 ${CAT} ${cfile}.cmp | ${TR} 'edD' 'E' | ${BASE_DIR}/myBC.x > ./diff.err 2>> ${OUT_DIR}/${SUPPRESS}

 if [ ${?} -ne 0 ]; then
    append_diff "${ckey}|${kword}|${fstring}" errors 
    return 1
 fi
 return 0
}


append_diff ()
{
local strng="${1}"
local file="${2}"
string_separator "Check ${strng}" "-" >> ${file}
${CAT} diff.err >> ${file}
string_separator "END ${strng}" "-" >> ${file}     
echo >> ${file}
}
###################################################################################
#             Compare two files  either literally or numerically                  #
###################################################################################
 Compare_Files()
{
# $1 file to test
# $2 ref_file
# $3 key
# $4 keyword
# $5 filestring
# $6 keytype
cfile="${1}"
rfile="${2}"
ckey="${3}"
kword="${4}"
fstring="${5}"
ktype="${6}"
 Files_Exist ${cfile} ${rfile}
 if (( ${?} )); then return 1; fi

 case ${ktype} in
   INFO)
    ${DIFF} ${cfile} ${rfile} > ./diff.err  
    append_diff "${ckey}|${kword}|${cfile}" info
    return 0
    ;;
   LIT*)
    ${DIFF} ${cfile} ${rfile} > ./diff.err
    if [ -s "diff.err" ]; then 
        append_diff "${ckey}|${kword}|${cfile}" errors
        return 1
     fi
    ;;
   NUM*)
        Extract_Numbers ${cfile} ${rfile}
        Compare_Files_Num ${cfile} ${rfile} ${ckey} ${kword} ${fstring} ${ktype}
        if (( ${?} )); then return 1; fi
    ;;
    *) errecho "Unknown type \"${ktype}\" for keyword ${kword}"
       return 1
    ;;
  esac

 return 0
}

###################################################################################
#                     Analysis of the particular keyword                          #
###################################################################################
 analyse_keyword()
{
 line="${1}"
 snum="${2}"
 local err=false
 if [ -z "${line}" ]; then 
   errecho "   analyse_keyword: empty line"
   return 1;
 fi
 
 keyword=$(GetField 1 "${line}")
     key=$(GetField 2 "${line}")
 keytype=$(GetField 3 "${line}")
   kfile=$(GetField 4 "${line}")
 case ${key} in
    CHECKSUM) file_ref="${kfile}.md5.${LEVEL}"
    ;;
    *)        file_ref="${kfile}.${REF}" 
    ;;
 esac
#prepend run number
 file="${snum}.${kfile}"
# all files are tested for existence
 Files_Exist ${file} ${file_ref}
 if (( ${?} )); then return 1; fi;
case ${key} in
  ABSSECTION|ABSVALUE|ABSCOLUMN)
        absfilter=strip_minus
        ;;
  *)    absfilter="${CAT} -"
        ;;
esac
case ${key} in
  FILE  ) Compare_Files ${file} ${file_ref} ${key} ${keyword} ${kfile} ${keytype};
          if (( ${?} )); then return 1; fi;;
 # gets all sections starting with headkey and ending with endkey
  SECTION|ABSSECTION) 
          headkey=$(GetField 5 "${line}")
           endkey=$(GetField 6 "${line}")
             mode=$(GetField 7 "${line}")
          case ${mode} in
             LAST)    ON=-1
                      OFF=-1
                      ;;
            [0-9]*-[0-9]*) ON=$(GetField 1 "${mode}" '-')
                          OFF=$(GetField 2 "${mode}" '-')
                      ;;
             [0-9]* ) ON=${mode}
                      OFF=${ON};
                      ;;
             FIRST)   ON=1
                      OFF=${ON}
                      ;;
             ALL|* )  ON=1;
                      OFF=-1;
                      ;;
          esac;
          ${AWK} "/${headkey}/ {if(waittoend==0){waittoend=1; mystart++;i=0; found=NR;\
                    if(mystart >= $ON && (mystart<= $OFF || $OFF < 0) ) doprint=1;}};\
                    {if(doprint>0){ if($ON>0){ print }else{ a[i++]=\$0 }}};\
                    /${endkey}/ {if(found!=NR) {doprint=0; waittoend=0}} ;\
                    END {if($ON <0) for(j=0;j<i;j++) print a[j]}" \
                    ${file} |\
                    ${absfilter} > ${snum}.${key}.${keyword} 2>> ${OUT_DIR}/${SUPPRESS};
          ${AWK} "/${headkey}/  {if(waittoend==0){waittoend=1; mystart++;i=0; found=NR;\
                                if(mystart >= $ON && (mystart<= $OFF || $OFF < 0) ) doprint=1;}};\
                               {if(doprint>0){ if($ON>0){ print }else{ a[i++]=\$0 }}};\
                  /${endkey}/  {if(found!=NR) {doprint=0; waittoend=0}} ;\
                  END          {if($ON <0) for(j=0;j<i;j++) print a[j]}" \
                  ${file_ref} |\
                  ${absfilter} > ${snum}.${key}.${keyword}.${REF} 2>> ${OUT_DIR}/${SUPPRESS};
          Compare_Files ${snum}.${key}.${keyword} ${snum}.${key}.${keyword}.${REF} ${key} ${keyword} ${file} ${keytype};
          if (( ${?} )); then 
             return 1; 
          fi
          ;;
  VALUE|ABSVALUE ) 
          value=$(GetField 5 "${line}")
            pos=$(GetField 6 "${line}")
           mode=$(GetField 7 "${line}")
          case ${mode} in
             ALL|* )   filter="${CAT} -";;
     [0-9]|[0-9][0-9]) filter="${SED} -n \"${mode} p\"";;
             LAST )    filter="${TAIL} -1";;
             FIRST   ) filter="${HEAD} -1";;
          esac
          ${AWK}  "/${value}/ {print \$${pos}}" ${file} | ${filter} | ${absfilter} \
                     >   ${snum}.${key}.${keyword}         2>> ${OUT_DIR}/${SUPPRESS};
          if [ ! -s "${snum}.${key}.${keyword}" ]; then 
            echo "    Value '${value}' is not found in file: ${file} !!!" >> ./diff.err;
            err=true
          fi;
          ${AWK}  "/${value}/ {print \$${pos}}" ${file_ref} | ${filter} | ${absfilter} \
                     >   ${snum}.${key}.${keyword}.${REF}  2>> ${OUT_DIR}/${SUPPRESS};         
          if [ ! -s "${snum}.${key}.${keyword}.${REF}" ]; then 
            echo "    Value '${value}' is not found in file: ${file_ref} !!!" >> ./diff.err;
            err=true
          fi;
          if ${err} ; then
              append_diff "${keyword}|${key}|${file}" errors
              return 1
          fi
          Compare_Files ${snum}.${key}.${keyword} ${snum}.${key}.${keyword}.${REF} ${key} ${keyword} ${file} ${keytype}; 
          if (( ${?} )); then 
             return 1; 
          fi
          ;;
  COLUMN|ABSCOLUMN) pos=$(GetField 5 "${line}")
          ${AWK} "{print \$${pos};}" ${file}     | ${absfilter}  > ${snum}.${key}.${keyword}\
                  2>> ${OUT_DIR}/${SUPPRESS};
          ${AWK} "{print \$${pos};}" ${file_ref} | ${absfilter}  > ${snum}.${key}.${keyword}.${REF}\
                  2>> ${OUT_DIR}/${SUPPRESS};
          Compare_Files ${snum}.${key}.${keyword} ${snum}.${key}.${keyword}.${REF} ${key} ${keyword} ${file} ${keytype};
          if (( ${?} )); then 
             return 1 
          fi
          ;;
  SIZE  ) size=$(ls -l ${file} | ${AWK} '{print $5}');
          size_ref=$(ls -l ${file_ref} | ${AWK} '{print $5}');
          if [ -n "${size#${size_ref}}" ]; then  
            echo "Size of '${file}': ${size} \nSize of '${file_ref}': ${size_ref}" > diff.err           
            case ${keytype} in
                INFO)
                    append_diff "${keyword}|${key}|${file}" info                   
                    return 0
                ;;
                LIT*|NUM*)
                    append_diff "${keyword}|${key}|${file}" errors
                    return 1
                ;;
             esac
          fi
          ;;
CHECKSUM) ${MD5} --status  -c ${file_ref} 2> ./diff.err;
          if (( ${?} )); then
            case ${keytype} in
                INFO)
                    append_diff "${keyword}|${key}|${file}" info
                    return 0
                    ;;
                LIT*|NUM*)
                    append_diff "${keyword}|${key}|${file}" errors
                    return 1;
                    ;;
            esac
          fi
          ;;
       *) errecho "    Wrong ${key} !"
          return 2;
          ;;
 esac
 return 0
}

############################################################################
#### parse keywords.cfg and replace numeric setting according to local  ####
#### definition, numeric and accur option                               ####
############################################################################
 write_keyword()
{

# the keyword processed
mykey=${1}
# the type in the local config file if set
if ${locdef} ; then
    mytype=${2:-XXX}
else
    mytype="XXX"
fi
# the type in the global keywords.cfg
keytype=$(${AWK} "BEGIN {FS=\"|\"} \$1==\"${mykey}\" {print \$3}"    ${BASE_DIR}/keywords.cfg )

# replace type by local value if needed
case ${keytype} in
       INFO*) 
       # INFO won't be overridden
          outtype="INFO"
          ;;
       LIT*|NUM*)
          case ${mytype} in
             XXX)
               # $2 was empty
               # nothing is overridden
               outtype="${keytype}"
               ;;
             INFO|LIT|NUM*)
               # replace by local value
               outtype="${mytype}" 
               ;;
             *)
               # local type unknown
               errecho "Unknown local type \"${mytype}\" for keyword ${mykey}"
               errecho "    -> using \"${keytype}\" from keywords.cfg"
             ;;
          esac
           ;;
        *)
         # default type unknown
          errecho "Unknown default type \"${keytype}\" "
          errecho "  or unknown keyword \"${mykey}\" -> skipping"
          return 1
         ;;         
esac

# check "numeric" cmdline option
if [ ${numeric:-XXX} == "numeric" -a  "${outtype}" == "LIT" ] ; then
    # numeric and  literal comparison is set -> replace
    outtype="NUM"
fi 

# check "literal" cmdline option
if [ ${literal:-XXX} == "literal" -a "${outtype#NUM}" != "${outtype}" ] ; then
   # outtype is numeric -> replace by literal
   outtype="LIT"
fi

# process accurarcy issues
let mydigits=0
let myaccur=0
let outdigits=maxdigits

if [ "${outtype#NUM}" != "${outtype}" ] ; then 
   # we have a numerical test
   val=$(echo ${outtype} |${TR} -cd "+\-0123456789") 
   mydigits=$(echo ${val} |${AWK} -F\- '{print $1}')
   myaccur=$(echo ${val} |${AWK} -F\- '{print $2+0}')

   if [ ${mydigits} ] ; then
       if [ ${mydigits} -le ${outdigits} -o ${outdigits} -eq 0 ] ; then
           let   outdigits=mydigits
       fi
   fi
   let outaccur=myaccur+accur
   outtype="NUM${outdigits}-${outaccur}"
fi

# replace outtype in the keyword line
${AWK} "BEGIN {FS=\"|\"; OFS=FS } \$1==\"${mykey}\" {\$3=\"${outtype}\" ; print \$0}" \
   ${BASE_DIR}/keywords.cfg >> ./cfg.tmp  
return
}


###################################################################################
#                         Expand LIST keywords                                    #
###################################################################################
 expand_keyword()
{
  local keyword=${1}
  local level=${2}
  local myloctype=${3}
  local nextlevel
  if [ "${level}" == "5" ] ; then
    outerror "too many levels of nested keyword lists" >> ${OUT_DIR}/${ERR_LOG} 2> /dev/null
    end_test 11 "!!! loop suspected  in keyword list definitions !!!"     
  else 
    let nextlevel=level+1
  fi 
  keyfile="${BASE_DIR}/keywords.cfg"
  if [ ! -e "${keyfile}" ]; then 
    end_test 9 "File 'keywords.cfg' is not found in ${BASE_DIR} !"
  fi
  line=$(${AWK} "/^${keyword}\|/ {print} " ${keyfile})
  type=$(GetField 2 "${line}")
  if [ "${type}" == "LIST" ] ; then
        for newkey in $(echo ${line}|${AWK} -F\| '{for(i=3;i<=NF;i++)print $i}') ; do
            expand_keyword ${newkey} ${nextlevel} ${myloctype}
        done 
  else 
     write_keyword ${keyword} ${myloctype} >> ./errors 2>> ${OUT_DIR}/${ERR_LOG}
  fi
}

RunOK ()
{
local logfile=${1}
local crash=false
if [ ! -f ${logfile} ] ; then
   echo "PROGRAM logfile does not exists." > diff.err
   crash=true
else
    ${GREP} 'PROGRAM CPMD ENDED AT' ${logfile} >  /dev/null 
    if (( ${?} )); then
       echo "PROGRAM CRASHED. End of log file:" > diff.err
       ${TAIL} -5 ${logfile} >> diff.err 2>> ${OUT_DIR}/${SUPPRESS}
       crash=true
    fi
fi
if ${crash} ; then
    append_diff "PROGRAM TERMINATION" errors
    echo CRASHED
else
    echo PASSED
fi
}
###################################################################################
#                         Analysis of test results                                #
###################################################################################
 analyse_res()
{
 local result
 snum=${1}
 result=$(RunOK ${snum}.test.out)
 if [ ${result} != PASSED ] ; then
     echo ${result}
     return 1
 fi

 if [ -e test.cfg.${LEVEL} ]; then
   test -f ./cfg.tmp && ${RM} ./cfg.tmp
   while read line ; do 
# Dealing with extra spaces and empty lines
      line=$(echo ${line} | ${TR} -sd ' ' '')
      if [[ ${line} != "" ]]; then 
        key=$(GetField 1 "${line}")
        type=$(GetField 2 "${line}")
        expand_keyword ${key} 0  ${type}   2>> ${OUT_DIR}/${SUPPRESS}
      fi
   done < test.cfg.${LEVEL}
   ${SORT} ./cfg.tmp |${UNIQ} >./cfg  2>> ${OUT_DIR}/${SUPPRESS}
   ${RM} -rf ./cfg.tmp 
   REF="ref.${LEVEL}"
 else
   echo "    No configuration file is found!" >> ./errors
   echo FAILED
   return 1
 fi

 while read myline ; do
   analyse_keyword "${myline}" "${snum}"
   if (( ${?} )); then 
     result=FAILED
     #string_separator "TESTED '${myline}'"  '^' >> ./errors
     #string_separator "_"  '_' >> ./errors
   fi 
 done < cfg
 echo ${result}
 return 
}

###################################################################################
#                         print the summary of passed/failed tests                #
###################################################################################
Print_Summary()
{
 hline
 if ${add} ; then
   outecho "SUMMARY:  ${tot_failed} tests FAILED,  ${tot_skipped} tests SKIPPED,  ${tot_added} tests  ADDED "
 else
   outecho "SUMMARY:  ${tot_failed} tests FAILED,  ${tot_skipped} tests SKIPPED,  ${tot_passed} tests PASSED"
 fi
 hline
}

###################################################################################
#                      store reference files for a test                           #
###################################################################################
 add_test()
{
 local result
 snum=${1}
 result=$(RunOK ${snum}.test.out)

 if [ ${result} = CRASHED ] ; then
     echo CRASHED
     return
 fi
 for ref_file in $(${CAT} ${BASE_DIR}/${CPMD_FLIST}) ; do 
   ref=$(GetField 1 "${ref_file}")
   opt=$(GetField 2 "${ref_file}")
   for file in $(ls ${snum}.${ref} 2>/dev/null) ; do
     if [[ ${opt} != "MD5" ]]; then
       ${MV} ${file} ${file#${snum}\.}.ref.${LEVEL}
     else
       ${MD5} ${file} > ${file#${snum}\.}.md5.${LEVEL}
       ${RM}  ${file}
     fi
   done
 done
 echo ADDED
}

###################################################################################
#                    perform tests within the test directory                      #
###################################################################################
 make_test()
{
    testname=${1}
    let tag=1
    if [ -e "skipped_tests" ]; then 
       ${GREP} "${testname#./}" skipped_tests >> ${OUT_DIR}/${SUPPRESS} 2>&1
       if (( ${?} == 0 )); then
         outecho_column -3 "  ${testname##/}" 
         outecho_column 4 ""
         outecho_status "SKIPPED"
         let skipped=skipped+1
         let tag=0
       fi
    fi
    if [ ${tag} -ne 0 ]; then
      cd ${testname} 
      clean_suitefiles TEST
      TD=${PWD%/*}
      STR=${TD##*/}
      # do we have to generate a list of files to be ignored by the version control ?
      if ${vcig}  ; then
          VC_ignore
      fi
      outecho_column -3 "  ${testname##*/} _____________________________"
      # generate input files and check how many we have
      ni=$(gen_inputs)
       if [ ${ni} -eq 0 ] ; then 
         let skipped=skipped+1
         outecho_column -4 "no input"
         result="SKIPPED"
       else
         while  read inputs ; do
            runnum=$(GetField 1 "${inputs}")
            sstr="$(GetField 2 "${inputs}")"
            np=$(GetField 4 "${inputs}")
            if [ ${runnum} -gt 0 ] ; then
               outecho_column 3 "|__________________"
            fi
            outecho_column -3 "${runnum}: $(GetField 3 "${inputs}") ________________________________"
        
            if ${run} ; then
                clean_run ${runnum}
                cp_input_files "${sstr}"
                STARTT=$(myseconds)
                Run_CPMD ${PWD}/input ${PWD}/test.out ${np}
                ENDT=$(myseconds)
                Eltime ${STARTT} ${ENDT}
                CopyRunOutput "${runnum}"
            else outecho_column 1 ""
            fi
            if ${add} ; then
                result=$(add_test ${runnum})
            elif ${wipe} ; then
                clean_suitefiles TEST    ${runnum}
                clean_suitefiles KEYWORD ${runnum}
                clean_suitefiles RUN     ${runnum}
                clean_run  ${runnum}
                result=CLEANED    
            else
                result=$(analyse_res ${runnum})
                if [ -s ./info ] ; then  
                    echo "${PWD#${BASE_DIR}} at ${datetr}" >> ${OUT_DIR}/${INFO_LOG}
                    echo "--------------------------------------------------------------------------------" >> ${OUT_DIR}/${INFO_LOG}
                    ${CAT} ./info >> ${OUT_DIR}/${INFO_LOG} 2> /dev/null
                    echo "--------------------------------------------------------------------------------" >> ${OUT_DIR}/${INFO_LOG}
                fi
            fi
            # add and analyse are done; remove all CPMD output
            outecho_status "${result}"
            case ${result} in
                CRASHED|FAILED)
                    string_separator "FAILED TEST: ${PWD#${BASE_DIR}/}" '>'        >> ${OUT_DIR}/${ERR_LOG}
                    echo "VARIANT: $(GetField 3 "${inputs}")" >> ${OUT_DIR}/${ERR_LOG}

                    DesName=$(${FIND} . | grep -i test.ts)
                    string_separator "" "---DESCRIPTION---" >> ${OUT_DIR}/${ERR_LOG}
                    if [ -e "${DesName}" ]; then 
                        ${CAT} ${DesName}  >> ${OUT_DIR}/${ERR_LOG}
                        echo >> ${OUT_DIR}/${ERR_LOG}
                    else
                        echo "Description of the test is missing !!!"  >> ${OUT_DIR}/${ERR_LOG}
                    fi
                    string_separator "" "--CHECKS--" >> ${OUT_DIR}/${ERR_LOG}
                    ${CAT} ./errors >> ${OUT_DIR}/${ERR_LOG} 2> /dev/null
                    string_separator 'END' '<'       >> ${OUT_DIR}/${ERR_LOG}
                    echo >> ${OUT_DIR}/${ERR_LOG}
                    echo "${testname#./}" >> ${OUT_DIR}/file.temp
                    let failed=failed+1
                    ;;
                PASSED) 
                    let passed=passed+1
                    ;;
                ADDED) 
                    let added=added+1
                    clean_suitefiles TEST    ${runnum}
                    clean_suitefiles KEYWORD ${runnum}
                    clean_suitefiles RUN     ${runnum}
                    ;;
                SKIPPED)
                    let skipped=skipped+1
                    ;;
                CLEANED)
                    let wiped=wiped+1
                    ;;
                *)
                    end_test 1 "unknown test result"
                    ;;
            esac
        done < curtests
        fi
     cd ${BASE_DIR}
   fi
 return ${failed}
}

###################################################################################
#              Proceeding the test directory and performing tests                 #
###################################################################################
 ini_test_dir() 
{
 local mydir=${1}
 let added=0
 let passed=0
 let failed=0
 let skipped=0
 #check for description in the root directory
 if [ -f "${mydir}/Description.TS" ]; then 
   hline
   info="$(${CAT} ${mydir}/Description.TS)"
 else
   info=${mydir}
 fi 
 outecho_column -3  "${info}"
 outecho_column -3  "VARIANT"
 if ${run} ; then
   outecho_column -1  "TIME"
 else
   outecho_column -1  ""
 fi
 outecho_status "STATUS"
 return 
}

# save results of a parent directory
add_tot()
{
 let   tot_added=tot_added+added
 let  tot_passed=tot_passed+passed
 let tot_skipped=tot_skipped+skipped
 let  tot_failed=tot_failed+failed
}
###################################################################################
#                output the directory tests' results                              #
###################################################################################
 test_dir_res()
{
 local name=${1}
 let sum=passed+failed+skipped+added
 if [ ${sum} -gt 0 ]; then
   outecho_column 1 
   outecho_column -2 "FAILED ${failed}"
   outecho_column -2 "SKIPPED ${skipped}"
   if ${add} ; then
      outecho_column -1 "ADDED ${added}"
   else
      outecho_column -1 "PASSED ${passed}"
   fi
   outecho_column 2 "${name#\.\/}-SUMMARY"
   outecho
   hline
 fi
}

###################################################################################
#                     gather and print help message                               #
###################################################################################
 print_help() 
{
 echo ""
 echo "      This is a beta version of the CPMD Test Suite."
 echo "      Please check the tex-file Description.tex"
 echo "      in the Test Suite directory"
 echo "      for a complete manual."
 echo "      All bug reports, suggestions and comments should be sent to:"
 echo "      Sergei Ivanov: sergei.ivanov@theochem.ruhr-uni-bochum.de"
 echo "                                  or "
 echo "      Gerald Mathias: gerald.mathias@theochem.ruhr-uni-bochum.de"
 echo ""

 ${AWK} 'pr==1 {pr=0;cmd=substr($1,0,length($1)-1); cmd=cmd " " val;\
                printf("  -%-20s  %s\n",cmd,txt) }\
         $1=="#CMT" {pr=1; txt=$0; gsub("#CMT ","",txt); \
             if($2=="VALUE"){val="<val>"; gsub("VALUE ","",txt)} else{val=""} }' ${0}

 exit 11
}
###################################################################################
#                     gather PARENT test directories                              #
###################################################################################
SelDir ()
{
  ndir="${1}"
  if [ -z "${TESTDIRS}" ] ; then
    TESTDIRS="${ndir}"
  else
    # check if directory has already been selected
    for odir in ${TESTDIRS//|/ } ; do
        if [ "${odir}" = "${ndir}" ] ; then
           echo ${TESTDIRS}
           return
        fi
    done
    TESTDIRS="${TESTDIRS}|${ndir}"
  fi
  echo "${TESTDIRS}"

}

SelAllDir ()
{   
   if [ -z "${ALLTESTS}" ] ; then
      end_test 1 "variable ALLTESTS not set in the defaults"
   else
      for ndir in ${ALLTESTS//|/ } ;  do
        TESTDIRS=$(SelDir "${ndir}")
      done
   fi
   echo ${TESTDIRS}
}

# look for Test.TS in all subdirectories
FindTest()
{
local dir=${1}
local str
${FIND} ${dir} -name "T[eE][sS][tT].TS"|\
${SED} 's@/T[eE][sS][tT].TS@@g; s@^./@@'
}


###################################################################################
#                     gather tests from directories and files                     #
###################################################################################
BuildTestList ()
{
 local dir
 local filter

 # clean old list
 test -e "${TEST_LIST}.tmp" && ${RM} -rf  "${TEST_LIST}.tmp"
 test -e "${TEST_LIST}"     && ${RM} -rf  "${TEST_LIST}"
 if [ -n "${TESTDIRS}" ] ; then
    for dir in ${TESTDIRS//|/ } ; do
        test -d ${dir} || continue
        FindTest ${dir} >> ${TEST_LIST}.tmp
    done
 fi
 if [ -n "${TEST_FILES}" ] ; then
    for file in ${TEST_FILES//|/ } ; do
        test -f ${file} || continue
        for tdir in $(${CAT} ${file}) ; do
                if [ -d ${tdir} ] ; then
                   echo ${tdir%\./} >> ${TEST_LIST}.tmp
                else
                   errecho "Specified '${tdir}' in test file '${file}' does not exist > SKIPPED"
                fi
        done
    done
 fi 
 if [ -s   ${TEST_LIST}.tmp ] ; then
    if [ -n "${SKIP_FILES}" ] ; then
        ${SED} ' s@^./@@; s@/ *$@@; s@[/ ]*$@@; s@^@^@; s@$@$@' ${SKIP_FILES//|/ } > skip.tmp
        if [ -s skip.tmp ] ; then
            filter="${GREP} -v -f skip.tmp "
        else
            filter="${CAT} - "
        fi
    else
            filter="${CAT} - "
    fi
    ${SORT} -u ${TEST_LIST}.tmp | ${filter} > ${TEST_LIST}
fi

${RM} -rf  ${TEST_LIST}.tmp 

}

###################################################################################
#                     command line parser                                         #
###################################################################################
ParseCommandLine()
{
while [ ${#} -ge 1 ]; do
case ${1:1} in
#CMT run tests in all directories
  ALL)      TESTDIRS=$(SelAllDir)
             ;;
#CMT VALUE read tests from this file
FILE|file)   if [ -z "${TEST_FILES}" ] ; then
                TEST_FILES="${2}"
             else
                TEST_FILES="${TEST_FILES}|${2}";
             fi
             echo ${TEST_FILES}
             shift
             ;;
#CMT select one of the default cpmd tests
  BO|CP|GO|MTD|PI|QMMM|WFO|VIB|VAR) 
            TESTDIRS=$(SelDir "${1:1}")
            ;;
#CMT VALUE do all tests in this directory
  DIR|dir)  TESTDIRS=$(SelDir ${2})
            shift
            ;;
  SKIP|skip) if [ -z "${SKIP_FILES}" ] ; then
                SKIP_FILES="${2}"
             else
                SKIP_FILES="${SKIP_FILES}|${2}";
             fi
             shift
             ;;
#CMT VALUE the base directory
  basedir)  BASE_DIR=${2}
            shift
            ;;
#CMT VALUE path to the executable
  bindir)   EXE_DIR=${2}
            shift
            ;;
#CMT VALUE path to the source
  srcdir)   SOURCE=${2};
            shift
            ;;   
#CMT VALUE the directory to compile in 
  compdir)  COMPILE=${2}
            shift
            ;;
#CMT VALUE the output directory
  outdir)   OUT_DIR=${2}
            shift
            ;;
#CMT VALUE the PP_LIBRARY_PATH
  libdir)   PP_LIBRARY=${2}
            shift
            ;;
#CMT VALUE the name of the executable
    exe)    CPMD=${2}
            shift
            ;;
#CMT VALUE endianess to use for restarts {BIG|LITTLE|AUTO}
    endian) endian=${2}
            shift
            ;;
#CMT VALUE command to start cpmd in parallel
  mpicmd)   MPICMD=${2}
            shift
            ;;
#CMT VALUE maximum number of processors available for parallel testing
  maxproc)  MAXPROC=${2}
            shift
            ;;
#CMT VALUE set version control to cvs, svn, none
  vc)       vc=${2}
            shift
            ;;
#CMT do/don't do a version control 'update' in the source directory
  vcupd|novcupd)   vcupd="$(GetBool ${1})"
            ;;
#CMT VALUE version control options string i.e. date, tag or revision number for a specific update
  vcopt)    vcopt=${2}
            shift
            ;;
#CMT do/don't build ignore files for the version control system in the test directories
  vcignore|novcignore)   vcig="$(GetBool ${1})"
            ;;
#CMT do/don't 'make clean' before compiling
  clean|noclean)    clean="$(GetBool ${1})"
            ;;
#CMT do a 'make reconfig' in the source directory 
  reconfig|noreconfig) reconf="$(GetBool ${1})"
            ;;
#CMT do/don't compile executable
  compile|nocompile)   compile="$(GetBool ${1})"
            ;;
#CMT VALUE use these options for make, e.g. "-j2 -f Makefile.pgi" 
  makeopt) makeopt="${2}"
           shift
           ;;
#CMT do/don't run executable (e.g. just compare the output)
  run|norun)    run="$(GetBool ${1})"
           ;;
#CMT generate/overwrite reference output files; no tests are performed
  add|noadd)  add="$(GetBool ${1})"
           ;;
#CMT clean test directories; implies norun
  wipe)  wipe="$(GetBool ${1})"
           ;;
#CMT VALUE perform test of this level
  level)    LEVEL=${2};
            shift
            ;;
#CMT replaces literal by numeric comparisons
  numeric) numeric="numeric"
            ;;
#CMT replaces numeric by literal comparisons
  literal) literal="literal"
            ;;
#CMT VALUE accuracy for numerical comparisons (e.g. 1 ignores the last digit)
  accur)   accur=${2};
           shift
           ;;
#CMT do/don't use local keyword definitions in tests
  locdef|nolocdef) locdef="$(GetBool ${1})"
           ;;
#CMT VALUE the maximum number of significant digits
  maxdig|maxdigits)  maxdigits=${2};
           shift
           ;;
#CMT this help
  h|help) print_help
            ;;

#  -script)   RUN_SCRIPT=${2};
#             shift;;
  *)         echo "unknown option ${1}";
             print_help
             exit 9
            ;;
esac
shift
done
}

###################################################################################
#                   set defaults for main directories                             #
###################################################################################
SetDirectories()
{
if [ -z "${BASE_DIR}" ];   then BASE_DIR=${PWD}                 ; fi
if [ -z "${SOURCE}"   ];   then   SOURCE=${BASE_DIR}/../SOURCE  ; fi
if [ -z "${OUT_DIR}"  ];   then  OUT_DIR=${BASE_DIR}/TS.OUT     ; fi
if [ -z "${PS_DIR}"   ];   then   PS_DIR=${BASE_DIR}/PP_LIBRARY ; fi
if [ -z "${COMPILE}"  ];   then  COMPILE=${SOURCE}              ; fi
if [ -z "${EXE_DIR}"  ];   then  EXE_DIR=${COMPILE}             ; fi

Dir_Exist "${OUT_DIR}" CREATE "Test Output"
OUT_DIR="$(absolute_path ${OUT_DIR})"
}


###################################################################################
###################################################################################
##                                            _                                  ##
##                            _ __ ___   __ _(_)_ __                             ##
##                           | '_ ` _ \ / _` | | '_ \                            ##
##                           | | | | | | (_| | | | | |                           ##
##                           |_| |_| |_|\__,_|_|_| |_|                           ##
##                                                                               ##
###################################################################################
###################################################################################

###################################################################################
#                   Initial settings of variables etc...                          #
###################################################################################

###################################################################################
#                       Defaults and command line arguments                       #
###################################################################################
if [ -n "TS.temp" ]; then 
    ${RM} -rf TS.temp
fi

# Load defaults
Load_Defaults

ParseCommandLine "${@}"
SetDirectories
Check_Input

###################################################################################
#                               backup and setup log files                        #
###################################################################################
echo
# This is done to avoid phasing out of the failed tests' file, before being read.
if [ -e ${OUT_DIR}/${FAIL_LIST} ]; then
  ${CP} ${OUT_DIR}/${FAIL_LIST} ${OUT_DIR}/f.tmp
fi

for log in ${ERR_LOG} ${OUT_LOG} ${MAKE_LOG} ${INFO_LOG} ${SUPPRESS} ${FAIL_LIST} ; do
   rotate_log ${log}
done
if [ -e ${OUT_DIR}/f.tmp ]; then 
  ${MV} ${OUT_DIR}/f.tmp ${OUT_DIR}/${FAIL_LIST} 2>> ${OUT_DIR}/${SUPPRESS}
fi

###################################################################################
#                               the numerics checker                              #
###################################################################################
make_myBC

###################################################################################
#                               start testing                                     #
###################################################################################
outecho ""
hline
hline
outecho_column -5 "Testing started" 
outecho_column 3 "$(date)"
outecho
hline

###################################################################################
#                            Check main directories                               #
###################################################################################
hline
outecho "Directories"

BASE_DIR="$(absolute_path ${BASE_DIR})"
Dir_Exist ${BASE_DIR} EXIT  "Test BASE"
Dir_Exist ${OUT_DIR} EXIT "Test Output"

if ${vcupd} ; then
  SOURCE="$(absolute_path ${SOURCE})"
  Dir_Exist ${SOURCE} EXIT    "CPMD Source"
fi
if ${compile} ; then 
  COMPILE="$(absolute_path ${COMPILE})"
  Dir_Exist ${COMPILE} EXIT   "Compilation"
fi
if ${run} ; then
  EXE_DIR="$(absolute_path ${EXE_DIR})"
  Dir_Exist ${EXE_DIR} CREATE "CPMD Executable"
  PS_DIR="$(absolute_path ${PS_DIR})"
  Dir_Exist ${PS_DIR} EXIT    "Pseudopotentials"
fi
hline

###################################################################################
#                Check binary encoding; Print test level                          #
###################################################################################

if [ "${endian}" == "AUTO" ] ;  then
   myendian=$(get_endian)
else
   myendian="${endian}"
fi
outecho_column -6 "Binary encoding used for restarts is"
outecho_column 2 "${myendian} ENDIAN"
outecho

outecho_column -7  "Level of testing"
outecho_status   ${LEVEL}
if [ ! -e "${BASE_DIR}/keywords.cfg" ]; then 
  end_test 9 "File 'keywords.cfg' is not found in ${BASE_DIR} !"
fi

###################################################################################
#                      Update the Source code if needed                           #
###################################################################################

outecho_column -7  "Version control <${vc}>: Updating SOURCE" 
if ${vcupd} ; then 
    cd ${SOURCE}
    string_separator "${vc} update ${vsopt}" >> ${OUT_DIR}/${MAKE_LOG}  2>&1
    ${vc} update ${vsopt} >> ${OUT_DIR}/${MAKE_LOG}   2>&1
    if (( ${?} )); then
      outecho_status "FAILED"
      end_test 1 "!!!  ${vc} update failed. Check ${OUT_DIR}/${MAKE_LOG}  !!!"
    else  
      outecho_status "FINISHED"
    fi
else
    outecho_status   "SKIPPED"
fi

###################################################################################
#                              Build the executable                               #
###################################################################################

outecho_column -7 "Compilation of the source"
if ${compile} ; then 
  cd ${COMPILE}
  if  ${clean} ; then
      string_separator "make clean" >> ${OUT_DIR}/${MAKE_LOG}  2>&1
      ${MAKE} ${makeopt} clean >> ${OUT_DIR}/${MAKE_LOG}  2>&1
      res=${?}
      if (( ${res} )); then
         outecho_status "FAILED"
         end_test 2 "!!! make clean failed. Check ${OUT_DIR}/${MAKE_LOG} !!!"
      fi
  fi    
  if ${reconf} ; then
       string_separator "make reconfig" >> ${OUT_DIR}/${MAKE_LOG}  2>&1
      ${MAKE} ${makeopt} reconfig >> ${OUT_DIR}/${MAKE_LOG} 2>&1
      res=${?}    
      if (( ${res} )); then
         outecho_status "FAILED"
         end_test 2 "!!! make reconfig failed. Check ${OUT_DIR}/${MAKE_LOG}  !!!"
      fi
  fi    
  string_separator "make" >> ${OUT_DIR}/${MAKE_LOG}  2>&1
  ${MAKE} ${makeopt} >> ${OUT_DIR}/${MAKE_LOG} 2>&1
  res=${?}
  string_separator "make finished" >> ${OUT_DIR}/${MAKE_LOG}  2>&1
  if (( ${res} )); then
    outecho_status "FAILED"
    end_test 2 "!!!  Compilation failed. Check ${OUT_DIR}/${MAKE_LOG}   !!!"
  else    
    outecho_status "FINISHED"
  fi
  cd ${BASE_DIR}
else
    outecho_status   "SKIPPED"
fi

###################################################################################
#                               Perform tests                                     #
###################################################################################
# Setting Tests' directories
tot_added=0
tot_passed=0
tot_skipped=0
tot_failed=0

BuildTestList

if [ -s ${TEST_LIST} ] ; then
  first=true
  touch ${OUT_DIR}/file.temp
  while read dir ; do
    root=${dir%/*}
    name=${dir##*/}
    if ${first} ; then
        oldroot=${root}
        ini_test_dir ${root}
        first=false
    fi
    if [ "${root}" != "${oldroot}" ] ; then
        add_tot
        test_dir_res ${oldroot}
        ini_test_dir ${root}
        oldroot=${root}
    fi
    make_test ${dir}
  done < ${TEST_LIST}
  add_tot
  test_dir_res ${oldroot}
  ${MV} ${OUT_DIR}/file.temp ${OUT_DIR}/${TEST_LIST}
  Print_Summary
  end_test 0 -n
else
  end_test 1 "no tests to process"        
fi
fi

###################################################################################
#                       _   _                           _                         #
#                      | |_| |__   ___    ___ _ __   __| |                        #
#                      | __| '_ \ / _ \  / _ \ '_ \ / _` |                        #
#                      | |_| | | |  __/ |  __/ | | | (_| |                        #
#                       \__|_| |_|\___|  \___|_| |_|\__,_|                        #
#                                                                                 #
###################################################################################
